// #################################################################################################
// # << NEORV32 - "Hello World" Demo Program >>                                                    #
// # ********************************************************************************************* #
// # BSD 3-Clause License                                                                          #
// #                                                                                               #
// # Copyright (c) 2021, Stephan Nolting. All rights reserved.                                     #
// #                                                                                               #
// # Redistribution and use in source and binary forms, with or without modification, are          #
// # permitted provided that the following conditions are met:                                     #
// #                                                                                               #
// # 1. Redistributions of source code must retain the above copyright notice, this list of        #
// #    conditions and the following disclaimer.                                                   #
// #                                                                                               #
// # 2. Redistributions in binary form must reproduce the above copyright notice, this list of     #
// #    conditions and the following disclaimer in the documentation and/or other materials        #
// #    provided with the distribution.                                                            #
// #                                                                                               #
// # 3. Neither the name of the copyright holder nor the names of its contributors may be used to  #
// #    endorse or promote products derived from this software without specific prior written      #
// #    permission.                                                                                #
// #                                                                                               #
// # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS   #
// # OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF               #
// # MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE    #
// # COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,     #
// # EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE #
// # GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    #
// # AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING     #
// # NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED  #
// # OF THE POSSIBILITY OF SUCH DAMAGE.                                                            #
// # ********************************************************************************************* #
// # The NEORV32 Processor - https://github.com/stnolting/neorv32              (c) Stephan Nolting #
// #################################################################################################
/*
 * This C program can multiply any two square or rectangular matrices.
 * The below program multiplies two square matrices of size 4 * 4.
 * There is also an example of a rectangular matrix for the same code (commented below).
 * We can change the Matrix value with the number of rows and columns (from MACROs) for Matrix-1
 * and Matrix-2 for different dimensions.
 */
 
/*
 * Note:  i- The number of columns in Matrix-1 must be equal to the number of rows in Matrix-2.
 *       ii- Output of multiplication of Matrix-1 and Matrix-2, results with equal to the number
 *           of rows of Matrix-1 and the number of columns of Matrix-2 i.e. rslt[R1][C2].
 */


/**********************************************************************//**
 * @file mmult/main.c
 * @author Stephan Nolting
 * @brief Classic 'matrix multiplication' demo program.
 **************************************************************************/

#include <neorv32.h>
#include <stdio.h>
#include <stdlib.h>


/**********************************************************************//**
 * @name User configuration
 **************************************************************************/
/**@{*/
/** UART BAUD rate */
#define BAUD_RATE 19200
/**@}*/

// Edit MACROs here, according to your Matrix Dimensions for mat1[R1][C1] and mat2[R2][C2]
#define R1 16            // number of rows in Matrix-1
#define C1 16            // number of columns in Matrix-1
#define R2 16            // number of rows in Matrix-2
#define C2 16            // number of columns in Matrix-2

/**********************************************************************//**
 * @name Function Prototypes
 **************************************************************************/

void mulMat(unsigned int mat1[][C1], unsigned int mat2[][C2]);

/**********************************************************************//**
 * Main function; prints some fancy stuff via UART.
 *
 * @note This program requires the UART interface to be synthesized.
 *
 * @return 0 if execution was successful
 **************************************************************************/
int main() {

        // capture all exceptions and give debug info via UART
        // this is not required, but keeps us safe
        neorv32_rte_setup();

        // init UART at default baud rate, no parity bits, ho hw flow control
        neorv32_uart0_setup(BAUD_RATE, PARITY_NONE, FLOW_CONTROL_NONE);

        // check available hardware extensions and compare with compiler flags
        neorv32_rte_check_isa(0); // silent = 0 -> show message if isa mismatch

        // print project logo via UART
        neorv32_rte_print_logo();

        // Square Matrices
        // R1 = 32, C1 = 32 and R2 = 32, C2 = 32 (Update these values in MACROs)
        unsigned int mat1[R1][C1]; 
        for (int i = 0; i < R1; i++){
                for (int j = 0; j < C1; j++){
                        mat1[i][j] = (i+1) *100;
                }
        }

        // = {
        //         {1, 1, 1, 1},
        //         {2, 2, 2, 2},
        //         {3, 3, 3, 3},
        //         {4, 4, 4, 4}
        // };

        unsigned int mat2[R2][C2];
        for (int i = 0; i < R2; i++){
                for (int j = 0; j < C2; j++){
                        mat2[i][j] = (i+1) *100;
                }
        }

        //  = {
        //         {1, 1, 1, 1},
        //         {2, 2, 2, 2},
        //         {3, 3, 3, 3},
        //         {4, 4, 4, 4}
        // };

        /*
        // Rectangular Matrices
        // R1 = 3, C1 = 4 and R2 = 4, C2 = 3 (Update these values in MACROs)
        int mat1[R1][C1] = {
                {1, 1, 1, 1},
                {2, 2, 2, 2},
                {3, 3, 3, 3}
        };

        int mat2[R2][C2] = {
                {1, 1, 1},
                {2, 2, 2},
                {3, 3, 3},
                {4, 4, 4}
        };
        */

        if (C1 != R2) {
        neorv32_uart0_print("The number of columns in Matrix-1  must be equal to the number of rows in "
                "Matrix-2\n");
        neorv32_uart0_print("Please update MACROs value according to your array dimension in "
                "#define section\n");

        exit(EXIT_FAILURE);
        }
        
        while(1) {
                char rcv_char = neorv32_uart0_getc();
                //neorv32_uart0_printf("%c\n", rcv_char);
                if (rcv_char == 'X') {
                        mulMat(mat1, mat2);
                }
        }
        
        return 0;
}


void mulMat(unsigned int mat1[][C1], unsigned int mat2[][C2]) {
    int rslt[R1][C2];
 
    //neorv32_uart0_print("Multiplication of given two matrices is:\n\n");
 
    for (int i = 0; i < R1; i++) {
        for (int j = 0; j < C2; j++) {
            rslt[i][j] = 0;
 
            for (int k = 0; k < R2; k++) {
                rslt[i][j] += mat1[i][k] * mat2[k][j];
            }
 
            //neorv32_uart0_printf("%d\t", rslt[i][j]);
            neorv32_uart0_printf("%d ", rslt[i][j]);
        }
        //neorv32_uart0_print("\n");
    }
    neorv32_uart0_print("\n");
}

 